function mrg = mergeSingleNeighbour(bN, has_src, is_ext)
%Merge block with single neighbour into that neighbour
%
% SYNOPSIS:
%   mrg = mergeSingleNeighbour(bN, has_src, is_ext)
%
% PARAMETERS:
%   bN - Block neighbourship as defined by function blockNeighbours.
%
%   has_src -
%        Logical flag that identifies blocks directly affected by source
%        terms (e.g., injection wells).  Blocks with source terms will not
%        be merged into their neighbours even if they have but a single
%        neighbour.
%
%   is_ext -
%        Logical flag that identifies blocks that are on the model
%        boundary.  These blocks will not be merged into their neighbours
%        even if they have but a single neighbour.
%
% RETURNS:
%   mrg - Merge operator that represents the mapping of original block IDs
%         to new block IDs.  The statement
%
%            q = mrg(p)
%
%         constructs a new partition vector, q, from an original partition
%         vector p in such a way that single-neighbour blocks of 'p' are
%         merged into their neighbours in 'q'.
%
% SEE ALSO:
%   mergeBlocks2.

%{
Copyright 2009-2015 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

   assert (~any(any(bN < 1)), ...
          ['Neighbourship definition must include ', ...
           'internal connections only.']);

   conn   = blockConnectivity(bN);
   nneigh = cellfun('prodofsize', conn);

   mrg = (1 : numel(has_src)) .';

   has_one_neigh = nneigh == 1;
   has_force     = has_src | is_ext;

   need_merge = find(has_one_neigh & ~has_force);

   while ~isempty(need_merge),
      next_merge = [];

      for b = reshape(need_merge, 1, []),
         [mrg, next_merge, conn] = merge_block(mrg, next_merge, b, conn);
      end

      need_merge = next_merge(~has_force(next_merge));
   end
end

%--------------------------------------------------------------------------

function [mrg, next, conn] = merge_block(mrg, next, b, conn)
   n = blockNeighbours(conn, b);
   n = unique(mrg(n));

   assert (numel(n) == 1, 'Internal error defining block neighbours.');

   mrg(mrg == b) = n;       % Merge 'b' into 'n'
   conn{b} = [];            % Quash the rebellion
   conn{n} = mrg(conn{n});  % Remove (now) undefined 'b' from n's neighs.

   u = unique(blockNeighbours(conn, n));
   if numel(u) == 1,
      next = [next, n];
   end
end
