classdef MultiscalePressureModel < ReservoirModel
    % Two phase oil/water system without dissolution
    properties
        pressureModel
        multiscaleSolver
        blockIndices
        useMex
        useReconstructedPressure
        coarseScaleConvergence
        reconstruct
    end
    
    methods
        
        function model = MultiscalePressureModel(G, rock, fluid, pmodel, mssolver, varargin)
            model = model@ReservoirModel(G, rock, fluid);

            model.useMex = false;
            model.pressureModel = pmodel;
            model.multiscaleSolver = mssolver;
            model.useReconstructedPressure = false;
            model.coarseScaleConvergence = false;
            model.reconstruct = true;
            
            model = merge_options(model, varargin{:});
            
            if model.useMex
                CG = mssolver.coarsegrid;
                mrstModule add mpfa
                tmp = accumarray(CG.partition, ones(G.cells.num, 1));
                [i, j] = blockDiagIndex(tmp);

                model.blockIndices = struct('I', i, 'J', j, 'blocksz', tmp);
            end
        end
        
        function [state, report] = stepFunction(model, state, state0, dt, drivingForces, linsolve, nonlinsolve, iteration, varargin)
%             [state, report] = model.pressureModel.stepFunction(state, state0, dt, drivingForces, model.multiscaleSolver, nonlinsolve, iteration, varargin{:});
            [state, report] = stepFunction@PhysicalModel(model, state, state0, dt, drivingForces, model.multiscaleSolver, nonlinsolve, iteration, varargin{:});
        end
        
        function [problem, state] = getEquations(model, state0, state, dt, drivingForces, varargin)
            [problem, state] = model.pressureModel.getEquations(state0, state, dt, drivingForces, varargin{:});
        end

        function [convergence, values, names] = checkConvergence(model, problem, varargin)
            [convergence, values, names] = checkConvergence(model.pressureModel, problem, varargin{:});
        end
        
        function [state, report] = updateAfterConvergence(model, state0, state, dt, forces)
            if model.reconstruct
                dispif(model.verbose, 'Pressure converged, reconstructing velocity field...');
                timer = tic();
                [state, report_ms] = model.reconstructFluxes(state0, state, dt, forces);
                dispif(model.verbose, ' Used %s \n', formatTimeRange(toc(timer))); 
            else
                report_ms = [];
            end
            if isfield(state, 'report')
                state = rmfield(state, 'report');
            end
            [state, report] = model.pressureModel.updateAfterConvergence(state0, state, dt, forces);
            report.reconstruction = report_ms;
        end
    
        function [state, varargout] = updateState(model, state, varargin)
            varargout = cell(1, nargout-1);
            p0 = state.pressure;
            [state, varargout{:}] = model.pressureModel.updateState(state, varargin{:});
            if model.coarseScaleConvergence
                part = model.multiscaleSolver.coarsegrid.partition;
%                 dp0 = state.dpRel;
%                 state.dpRel = accumarray(part, state.pressure - p0)./accumarray(part, p0);
                state.dpRel = accumarray(part, abs(state.dpRel))./accumarray(part, 1);
%                 fprintf('Average: %2.4g, Point: %2.4g\n', norm(state.dpRel, inf), norm(dp0, inf));
%                 c = model.multiscaleSolver.coarsegrid.cells.centers;
%                 state.dpRel = state.dpRel(c);
            end
        end
        
        function varargout = getVariableField(model, varargin)
            varargout = cell(1, nargout);
            [varargout{:}] = model.pressureModel.getVariableField(varargin{:});
        end
        
        % Utility functions 
        function [state, report] = reconstructFluxes(model, state0, stateConverged, dt, forces)
            CG = model.multiscaleSolver.coarsegrid;
            nc = CG.parent.cells.num;
            
            timer = tic();
            
            propPressure = stateConverged.pressure;
            
            problem = model.pressureModel.getEquations(state0, stateConverged, dt, forces,...
                'iteration', inf, 'staticwells', true);
            A = -problem.equations{1}.jac{1};
            b =  problem.equations{1}.val;
            
            % Single pass of multiscale solver
            solver = model.multiscaleSolver;
            its = solver.maxIterations;
            solver.maxIterations = 0;

            dpCoarse = solver.solveLinearSystem(A, b);
            
            solver.maxIterations = its;
            
            [dpFlux, t_solve] = reconstructPressure(CG.partition, dpCoarse, A, b);
            
            % Update with reconstruction and prolongation increments
            [stateCoarse, stateFlux] = deal(stateConverged);
            stateCoarse.pressure = propPressure  + dpCoarse(1:nc);
            stateFlux.pressure =  propPressure + dpFlux(1:nc);
            
            % Store fluxes and use it the corect places
            stateFlux   = setFluxes(model, state0, stateFlux,   dt, forces, propPressure);
            stateCoarse = setFluxes(model, state0, stateCoarse, dt, forces, propPressure);
            
            flux = stateFlux.flux;
            flux(CG.faces.fconn, :) = stateCoarse.flux(CG.faces.fconn, :);
            
            state = stateConverged;
            state.flux = flux;
            
            % Use property pressure for transport
            state.pressure = propPressure;

            report.reconstructionTime = toc(timer);
            report.reconstructionSolver = t_solve;
            if 0
                BHPix = find(arrayfun(@(x) strcmpi(x.type, 'bhp'), state.wellSol));
                for i = 1:numel(BHPix)
                    ix = BHPix(i);
%                     state.wellSol(ix) = stateFlux.wellSol(ix);
                    state.wellSol(ix).flux = stateFlux.wellSol(ix).flux;
                end
            end

        end
        
        function state_flux = setFluxes(model, state0, state, dt, forces, propsPressure)
            [~, state_flux] = model.pressureModel.getEquations(state0, state, dt, forces, ...
                                       'ResOnly', true, 'iteration', inf, ...
                                       'propsPressure', propsPressure, 'staticwells', true);

         end
    end
    

end
