function [state, model, schedule] = setupSPE10_AD(varargin)
%Create base case for 2nd model of 10th SPE Comparative Solutions Project
%
% SYNOPSIS:
%   [state, model, schedule] = setupSPE10_AD
%   [state, model, schedule] = setupSPE10_AD('pn1', pv1, ...)
%
% PARAMETERS:
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%
%               - layers --
%                   Vector of layer IDs from which to extract petrophyscial
%                   (rock) properties.  Default value: layers = 1:85 (all
%                   model layers).
%
%               - dt --
%                   Time step size.  Positive scalar.
%                   Default value: dt = 30*day.
%
%               - T --
%                   Total simulation scenario time.  Positive scalar.
%                   Default value: T = 2000*day.
%
%               - minporo --
%                   Minimum porosity value.  Non-negative scalar.  The
%                   rock's porosity field is adjusted according to the
%                   formula
%
%                      poro = MAX(rock, minporo)
%
%                   Default value: minporo = 0.01.
%
% RETURNS:
%   state    - Initial reservoir/model state defined on the specified
%              layers of the base case.
%
%   model    - The physical model for the SPE 10 base case.
%
%   schedule - Schedule defining the report steps for the base case.
%
% SEE ALSO:
%   simulateScheduleAD.

%{
Copyright 2009-2015 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

    opt = struct('layers',  1:85, ...
                 'dt',      30*day, ...
                 'T',       2000*day, ...
                 'minporo', 0.01);
    opt = merge_options(opt, varargin{:});
    
    mrstModule add spe10 ad-props ad-blackoil ad-core
    
    srw = 0.2;
    sro = 0.2;
    pRef = 6000*psia;
    
    % Fluid relative permeabilities
    fluid.krW = coreyPhaseRelpermAD(2, srw, 1, srw + sro);
    fluid.krO = coreyPhaseRelpermAD(2, sro, 1, srw + sro);
    
    % Water props
    bW0 = 1./(1.01);
    fluid.bW = @(p) bW0*exp((p - pRef)*3e-6/psia);
    fluid.muW = @(p) 0*p + 0.3*centi*poise;
    
    % Oil props
    p = [300; 800; 8000; 8001]*psia;
    b = 1./[1.05; 1.02; 1.01; 1.01];
    mu = [2.85; 2.99; 3; 3]*centi*poise;
    [fluid.bO, fluid.muO] = tableByPressureLinearAD(p, b, mu);
    
    
    fluid.rhoWS = 64*pound/(ft^3);
    fluid.rhoOS = 53*pound/(ft^3);

    % Rock compressibility
    cR = 1e-6/psia;
    fluid.cR = cR;
    fluid.pvMultR = @(p)(1 + cR.*(p-pRef));

    
    rock = SPE10_rock(opt.layers);
    rock.perm = convertFrom(rock.perm, milli*darcy);
    
    % Compute pore volume fraction of the full model
    volFrac = sum(rock.poro)/1.9141e+05;
    rock.poro(rock.poro < opt.minporo) = opt.minporo;
    
    
    % Grid
    cartDims = [  60,  220,   numel(opt.layers)];
    physDims = [1200, 2200, 2*cartDims(end)] .* ft();

    G = cartGrid(cartDims, physDims);
    try
        mrstModule add libgeometry
        G = mcomputeGeometry(G);
    catch
        G = mcomputeGeometry(G);
    end
    model = TwoPhaseOilWaterModel(G, rock, fluid, 'gravity', [0, 0, 9.80665]);
    model.minimumPressure = 1000*psia;
    
    state = initResSol(G, pRef, [srw, 1-srw]);
    
    % Wells
    makeProd = @(W, name, I, J) verticalWell(W, G, rock, I, J, [],...
        'Name', name, 'radius', 5*inch, 'sign', -1, 'Type', 'bhp',...
        'Val', 4000*psia, 'comp_i', [.5, .5]);
    W = [];
    W = makeProd(W, 'P1', 1, 1);
    W = makeProd(W, 'P2', 60, 1);
    W = makeProd(W, 'P3', 60, 220);
    W = makeProd(W, 'P4', 1, 220);
    W = verticalWell(W, G, rock, 30, 110, [], 'Name', 'I1', 'radius', 5*inch, ...
        'Type', 'rate', 'Val', volFrac*5000*stb/day, 'comp_i', [1, 0], 'Sign', 1);
    
    stepNum = opt.T/opt.dt;
    frac = rem(stepNum, 1);
    
    dt = repmat(opt.dt, floor(stepNum), 1);
    if frac > 0
        dt = [dt; opt.dt*frac];
    end
    
    schedule = simpleSchedule(dt, 'Wells', W);
end
